! $Id: ext2boot0.S,v 1.7 1994/07/03 15:06:35 root Exp $
!
! Package supporting Linux booting from Second Extended Filesystem.
! This file can be redistributed under the terms of the GNU General
! Public License
!
!			Copyright (c) 1994, Rafal Pietrak.
!

#include "ext2boot.h"

! new experimental boot loader for Linux
! Now, the plan is the following:
! - set SS and ES to somewhere in the upper region of real memory.
! - save current (our, boot) partition entries there.
! - get (and save) disk geometry
! - read remaining sectors in partition's fs first block (the actual sector
!   count will be a compilation parameter dependent on the fs block size --
!   for ext2 fs the block size is 2, so we must load (2 - 1) = 1 more sector)
! - compute offset to the current patrition (in flat sectors).
! - move everything up
!
! ld86 requires an entry symbol. This may as well be the usual one.
.globl	_main
_main:
!#if 0 /* hook for debugger, harmless unless BIOS is fussy (old HP) */
!	int	3
!#endif
!
! DS:SI point to current (i.e. the one we are loadded from) partition
! table entry.
!
	mov	ax, #INITSEG
	mov	es,ax
	mov	ss,ax
	mov	sp, #0x8000	! 0x8000 is arbitrary value >= length of
				! bootsect + length of setup + room for stack
	mov	bp, sp		! storage for local variables

#ifndef FLOPPY_BOOT		/* i.e. boot from hd */
!
! now, autodetect boot device.
! ... first manor number
!
	mov	cx, si
	sub	cl, #0x0be
	shr	cl, #4
	inc	cx
!
! ... then major number
!
	mov	al, #0x12
	out	#0x70, al
	in	al, #0xfe
	in	al, #0x71
	mov	ah, #3		! /dev/hd disk
	or	al, al
	jnz	from_hd
#ifdef XT_DISK_CONFIG
	mov	ah, #13		! /dev/xd disk
#else
	mov	ah, #8		! /dev/sd disk
#endif
from_hd:
	mov	al, cl
	mov	(bp+VAR_ROOT_DEV), ax

	lea	di, (bp+VAR_DISK)
	cld
	lodsw
	mov	al, #0x80	! make the partition 'active'
	stosw 			! save patrition's disk/head data...
	movsw 			! ...save patrition's sector/cylinder.

	mov	ah, #8			! AH=8 -> get drive parameters
	call	bios_io
					! DH contains total disk heads
					! DL = number of disks in system
#else				/* i.e. this is floppy boot */
	mov	ax, #0x201		! we'll read one sector
	mov	bx, #0x400		! there is some scratch area ... still.
	xor	cx, cx
	mov	(bp+VAR_DISK), cx	! set the 'load disk' to A:
	mov	si, #fd_data
fd_again:
	seg cs
	mov	cl, 0x7c00(si)
	or	cl, cl
	jnz	fd_doit
		lea	si, msg_diskio
		call	fatal_error
fd_doit:
	call	bios_io
	jnc	fd_ready
		inc	si
		jmp	fd_again
fd_ready:
	mov	(bp+VAR_ROOT_DEV), word #0x200
#endif
	push	es
	pop	ds			! DS = INITSEG

	and	ecx, #0x3f
	mov	(bp+VAR_SECTORS_TRACK), ecx	! save sectors/track
	xor	eax, eax
#ifdef FLOPPY_BOOT		/* floppies don't have more then two surfaces  */
	inc	ax			! 'last head' number for floppy is 1
#else
	mov	al, dh			! this is 'last head nr'
#endif
	inc	ax			! convert max_head to head_count
	mul	ecx
	mov	(bp+VAR_SECTORS_CYL), eax	! save sectors/cylinder
#ifndef FLOPPY_BOOT
	xchg	eax, ecx		! after: EAX == s/track, ECX == s/cylinder

	xor	ebx, ebx
	mov	bl, (bp+VAR_DISK+1)	! head of partition origin
	mul	ebx
	xchg	eax, ebx		! mov EAX -> EBX

!
! now compute the 'sector offset' to the partition origin
!
	mov	ax, (bp+VAR_SECT)	! get the origin of the partition
	push	ax			! +++ save VAR_SECT
	xchg	al, ah
! instead of " shr ah, 6", we use...
	db	0xc0, 0xec, 06		! compute the cylinder number of partition origin
!
	mul	ecx
	pop	cx			! --- restore VAR_SECT
	and	cx, #0x3f
	dec	cx
	add	eax, ebx
	add	eax, ecx		! *complete* *true* offset to linear 
#else
	xor	eax,eax			! for floppies there is only one partition 
#endif
	mov	(bp+VAR_PARTITION), eax	! sector of current partition begin.

	xor	eax, eax
	inc	ax			! (EAX == 1) -> first load super block
	mov	bx, #BUFF_BLOCK		! load into 'work area'
	call	load_block
	cmp	(bx+EXT2_MAGIC), #0xef53
	jnz	fatal_error
	mov	eax, (bx+EXT2_FIRST_DATA_BLOCK)	
	mov	(bp+VAR_FIRST_DATA_BLOCK), eax
	mov	eax, (bx+EXT2_BLOCKS_PER_GROUP)
	mov	(bp+VAR_BLOCKS_PER_GROUP), eax
	mov	eax, (bx+EXT2_INODES_PER_GROUP)
	mov	(bp+VAR_INODES_PER_GROUP), eax
!
! we won't need anything else from the superblock
!

!
! open BOOT inode
!
!	mov	eax, #EXT2_BOOT_LOADER_INO
	seg cs			! yes, this info is in the code segment (still)
	mov	eax, boot_inode_var+0x7c00	! yes, biased by 0x7c00 before
						! code is moved to #INITSEG
	call	get_inode
	jz	boot_load_failed ! continue loading system only if
				! get_inode didn't failed
	call	load_file	! loads an opened file @ predefined address (ES:0)
!
! the following code may be used only when fs-functions are controled
! by a jump table (in sector 0).
!
#ifdef WITH_UNWANTED
	push	cs
	pop	ds
	mov	cx, #0x100
	xor	di, di
	mov	si, #0x7c00
	rep			! overwrite code by bootsector
	movsw			! with this
#endif
	jmpi	#boot_input, #INITSEG
!
!
!**************************************************

boot_load_failed:
	mov	si, #msg_halt
!
! print error message pointed to by CS:SI, then halt the machine
!
fatal_error:
	push	cs
	pop	ds
	add	si, #0x7c00	! messages addess have to be adjusted
				! before code moves to the upper memory
	call	print_si
stop:	jmp	stop		! halt system

!
! EAX is a number of requested group
!
get_group:
	mov	bx, #BUFF_BLOCK
	db	0x66			! EAX will be multiplied
	mul	(bp+VAR_BLOCKS_PER_GROUP)
	add	eax, (bp+VAR_FIRST_DATA_BLOCK)
	mov	(bp+VAR_GROUP_ORIG_BLOCK), eax
	inc	eax
!
! fall through to LOAD_BLOCK....
!

! load fs logical block pointed to by EAX (makes BIOS translation), load
! is done at memory location ES:BX
!
load_block:
	shl	eax, #SIZEOF_FS_BLOCK_SHIFT
	add	eax, (bp+VAR_PARTITION)	! add offset to partition origin
	xor	edx, edx
	db	0x66			! 32-bit division
	div	(bp+VAR_SECTORS_CYL)	! (EAX == cylinder), (EDX == remainder)
! at this point we've got rid off the big numbers. sectors/heads will fit into
! 16BIT integer.
	push	ax
	mov	ax, dx
	xor	dx, dx
	div	(bp+VAR_SECTORS_TRACK)	! AX == head, DX == sector
	inc	dx			! sectors start from 1, not from 0
	pop	cx
	xchg	ch, cl
	ror	cl, #2			! rearange it the BIOS way
	or	cl, dl			! CX now ready for INT 13
	mov	dh, al
!
! load one filesystem block pointed by BIOS register set
!
bios_fs:
	mov	ax, #0x202		! BIOS_READ + FS_SECTORS_PER_BLOCK
!
! do a BIOS I/O refined by AX and other BIOS registers
!
bios_io:
	push	di
	mov	di, #5			! retry count
	push	si
	mov	si, #msg_diskio
rb_retry:
	mov	dl, (bp)		! always load from the same disk
	push	ax			! command saved to allow retry
	int	0x13
	jc	rb_reset
		pop	ax
		pop	si
		pop	di
		ret
rb_reset:
	sub	ax, ax
	int	0x13		! reset disk I/O subsystem
	pop	ax
	dec	di
	jnz	rb_retry
!	pop	si		! save one byte: don't cleanup stack ...
	jmp	fatal_error	! ... at this point it doesn't matter

!
! EAX is an inode describing the file
!
get_inode:
	or	eax, eax		! inode == 0 is invalid
	jnz	in_valid
		ret			! error return ZERO_FLAG == set
in_valid:
	mov	(bp+VAR_CUROPN_INODE), eax
	dec	eax			! fs structures map inode 1 to offset 0
	db	0x66
	xor	dx, dx
	db	0x66			! 32-bit opperation
	div	(bp+VAR_INODES_PER_GROUP) ! EAX = group, EDX = inode within group
	push	dx			! ok, the assumption is that IN_P_GR is less then 16bit
	call	get_group		! pointed to by EAX
	pop	ax			! which local inode
	xor	dx, dx
	mov	cx, #INODES_PER_BLOCK	! 8 == 1024 / 0x80
	div	cx			! AX = block whithin the group, DX == inode whithin block
	push	dx
	add	eax, (bx+EXT2GR_INODE_TABLE)
	add	eax, (bp+VAR_GROUP_ORIG_BLOCK)
	dec	eax			! I don't know why this is necesary 
	call	load_block
	pop	cx			! inode index in this block
	mov	ax, #SIZEOF_STRUCT_INODE ! sizeof(struct inode)
	mul	cx
	add	ax, bx			! the address is in the buffer pointed to by BX
	push	si
	push	di
	mov	si, ax
	lea	di, (bp+VAR_CURRENT_INODE)
	mov	cx, #SIZEOF_STRUCT_INODE
	rep				! yes, I could use "rep movsw" here, but
	movsb				! performance doesn't actualy matter at this point:)
	shr	(bp+VAR_CURRENT_INODE+EXT2IN_I_BLOCKS), #SIZEOF_FS_BLOCK_SHIFT
	pop	di
	pop	si
	or	bp, bp			! signal success
	ret
!
!
! locate EAX block of a file (current inode is valid)
!

file_block:
	push	si
	lea	si, (bp+VAR_CURRENT_INODE+EXT2IN_BLOCK)
!
! Doing the following shift opperation now (instead of doing it in each
! of the branches) limits the possible file size to 30 bits == 1Gbyte.
! Seems crude, but since we have only 640Mb of real memory...
! I do it to save one or two bytes in sector0 boot code.
! Again, the order of doing comparitions is chosen so, that only
! one (i.e. the first) comparition with 32-bit value is necesary.
!
	shl	eax, 2
	cmp	eax, #LOC_IND_2
	jc	fb_not3
		xor	eax, eax		! safeguard for unsupported indirection
		jmp	fb_exit
fb_not3:
	cmp	ax, #LOC_IND_1
	jc	fb_not2
		call	load_indirect1
		jmp	fb_exit
fb_not2:
	cmp	ax, #LOC_IND_0
	jc	fb_not1
		push	bx
		sub	ax, #LOC_IND_0
		push	ax		! save updated index
		mov	eax, (si+LOC_IND_0)
		call	load_indirect0
		pop	si
		mov	eax, (bx+si)	! this is location of block pointer
		pop	bx
		jmp	fb_exit
fb_not1:
	add	si, ax
	mov	eax, (si)
fb_exit:
	pop	si
	ret

!
! support routine. On return BX points to indirect cache.
!
load_indirect0:
	mov	bx, #BUFF_INODE
	cmp	eax, (bp+VAR_INODE_CACHE)
	jz	li_ret
	mov	(bp+VAR_INODE_CACHE), eax
	push	es
	push	ss
	pop	es
	call	load_block
	pop	es
li_ret:	ret

!
! load a file described by current inode (the opened one)
!
load_file:
	xor	eax, eax	! start loading from block 0
load_file_eax:
	push	es
next_block:
		push	eax
		call	file_block
		or	eax, eax
		jz	skip_block
			xor	bx, bx
			call	load_block
			mov	ax, #0x0e2e
! a little trickier print stuff, doesn't fit into 512 boot sector, though.
!			jmp	prnt_it
!			mov	ax, #0x0e2b
print_it:
			int	0x10
skip_block:
		mov	ax, es
		add	ax, #SIZEOF_FS_BLOCK_PAR
		mov	es, ax		! adjust load segment
		pop	eax
		inc	eax
	cmp	eax, (bp+VAR_CURRENT_INODE+EXT2IN_I_BLOCKS)
	jb	next_block
	push	eax		! just to preadjust the stack
end_load:
	pop	eax		! discard value (actually the number of blocks read)
	pop	es
	ret

print_si:
	lodsb
	or	al, al
	jnz	si_putc
		ret
si_putc: mov	ah, #0x0e
	int	0x10
	jmp	print_si

msg_diskio:
	.ascii	"Disk"
!	.ascii	" I/O error."
	.byte	0
msg_halt:
msg_nofs:
	.ascii	"No ext2fs."
	.byte	0

#ifdef FLOPPY_BOOT
fd_data:
	.byte	18, 15, 9, 8, 7, 0
	.byte	28,  0, 0, 0, 0, 0
#endif

.org 506
boot_inode_var:
	dd	EXT2_BOOT_LOADER_INO

.org 510
boot_flag:
	.word 0xAA55

	end
