! $Id: ext2boot1.S,v 1.8 1994/07/04 19:06:36 root Exp root $
!
! Package supporting Linux booting from Second Extended Filesystem.
! This file can be redistributed under the terms of the GNU General
! Public License
!
!			Copyright (c) 1994, Rafal Pietrak.
!

#include "ext2boot.h"

!
! the setup (i.e. boot) is loaded by now, so only the kernel proper
! still remains on disk. The trick is to select and load the kernel
! using its name.
!
	.align 256
boot_input:
!
! now, it's time to preserve some tratitional boot-sector data.
!
	mov	di, #0x1f0	! erase the parameter area...
	mov	cx, #6		! ... except for boot_from information
	xor	ax, ax
	rep
	stosw
!
!	mov	504, word #1500	! RAMDISK
!
! put actual root device
!
	mov	ax, (bp+VAR_ROOT_DEV)
	mov	508, ax

another_try:
!
! First, reset the kernel environment area. We want to start
! clean.
!
	mov	(bp+VAR_ENV_TAIL), word #BUFF_BOOT_ENV

	call	flush
	mov	si, #msg_boot
	call	print_si
	call	wait_for_key
	jz	use_default		! no key pressed for 10 seconds
	call	gets
!
! did we got any user input?
!
	lea	si, (bp+VAR_GETS_INPUT)
	cmp	(si), byte #0
	jnz	do_the_boot
	or	si, si
use_default:
		/*
		 * ZERO FLAG is set only if there was no
		 * keyboard input.
		 */
		jnz	not_auto
		mov	si, #msg_auto
		call	word_concat	! put into kernel ENV
not_auto:
		mov	si, #msg_image	! default kernel filename
do_the_boot:
!
! building kernel environment...
! ... which means reading command line in search for
! configuration file name. If found, useing it, otherwise
! we should use precompiled name: /etc/boot.conf.
! ... after reading the file, parse user input so that it
! takes precedence over the file contents. Append everything
! to BUFF_BOOT_ENV string (which already may contain the
! 'auto' string).
!
	call	build_kernel_env
	jc	another_try		! build wasn't succesful

!
! Look for some special enviroment variables (svga, bootflags,
! bootimage, etc...)
!
	call	lookup_bootflags
	call	lookup_svga
	call	lookup_ramdisk
!
! The following function *have* to locate kernel image name, copy
! it to some safe place where null terminated pathname could be 
! formed (i.e. into VAR_GETS_INPUT) and return SI pointing to it.
! Failing to do that, it should use some sensible default (creating
! propper ENV variable set to it).
! It may *not* return without something, so no error detection
! here.
!
	call	lookup_bootimage

!
! print some info...
!
	push	si
	mov	si, #msg_booting
	call	print_si
	pop	si
	push	si
	call	print_si
	pop	si
!
! ... and open kernel file once its name is known (and
! passed here in SI), we can load it then provided open
! was succesfull.
!
	call	open_root_path
	jnc	lookup_ok
!
! kernel filename lookup failed --> go try again...
!
		mov	si, #msg_nofile
		call	print_si
		jmp	another_try
!
! Ok, file is opened. We may load it into SYSSEG, now.
!
lookup_ok:
	mov	ax, #SYSSEG	! this is where the image gets loaded at
	mov	es, ax
	call	load_file	! now, load the kernel proper
	push	ss
	pop	es		! restore segment registers
!
! we could print some statistics about the file being loaded
! like its size, but...
!
	call	print_crlf

	call	kill_motor
!
! OK, this is it. We are through with the load bussiness; proceede to
! the actual setup code.
!
	call	setup_begin	! ...yup, I failed to make the as86 accept 'jmp' here

	jmp	another_try	! ... and this line is *never* reached.

!-------------------------------------------------------
!

!
! wait for a keypress 10 seconds. The cmos clock is used to get
! the time.
!
wait_for_key:
	call	gettime
	add	al,#10		! wait 10 seconds
	cmp	al,#60
	jl	lminute
	sub	al,#60
lminute:
	mov	cl,al
again:	mov	ah,#0x01
	int	0x16
	jnz	kw_exit		! key pressed, returning
	call	gettime
	cmp	al,cl
	jnz	again
kw_exit:
	ret


!
! Literaly: build environement for the kernel in the BUFF_BOOT_ENV.
! When it's not fruitfull to proceede (i.e. the help was requested)
! whith loading the kernel propper, set CARRY flag signaling the
! request to re-enter the main loop.
!
build_kernel_env:
no_spcs:
	lodsb
	cmp	al, #0x20
	jz	no_spcs

	dec	si
	call	strlen
	clc
	jcxz	pi_ret
	push	si
	cmp	(si), byte #'?
	jnz	no_help
!
! HELP requested; check what kind of help it'll be (printing
! the contents of help file or printing the directory).
!
		inc	si
		dec	cx
no_spcs2:
		lodsb
		cmp	al, #0x20
		loopz	no_spcs2
		dec	si
		jcxz	print_help
!
! There is something else in the gets input area; assume
! it's a directory name and print its contents.
!
		call	open_root_path
		jc	pi_exit
		call	print_dir
		jmp	pi_exit_restart
print_help:
!
! There should be a help file on this filesystem; locate
! it and then print.
!
		mov	si, #msg_help_file
		call	fopen
		jnc	next_line
			mov	si, #msg_nofile
			call	print_si
pi_exit_restart:
			stc
			jmp	pi_exit
next_line:
		call	fgets
		jc	pi_exit
		call	puts
		jmp	next_line
no_help:
!
! now locate the configuration file name (if any) in
! the input string.
!
#if 1
	call	lookup_config 		! look for the 'config=' string
#else
	mov	di, #msg_var_config 	! compare 'config=' string
	call	locate_string
	jz	nodefault
		mov	si, #msg_config
nodefault:
#endif
	call	fopen
	pop	si			! restore original pointer in case
	push	si
	jc	pi_no_file		! there is no configuration file
!
! save VAR_GETS_INPUT
! ok, this code breaks if the configuration file (/etc/boot.conf) is so
! large, that fgets will use second level of inode_block indirection.
!
		mov	di, #BUFF_INODE1
		mov	cx, #0x0d00	! this much at most
pi_copy:
		lodsb
		stosb
		or	al, al
		loopnz	pi_copy
config_loop:
		call	fgets
		jc	pi_file_end
		call	parse_fline
		jmp	config_loop
pi_file_end:
	mov	si, #BUFF_INODE1
	lea	di, (bp+VAR_GETS_INPUT)
	push	di
	mov	cx, #0x0d00		! this much at most
p2_copy:
	lodsb
	stosb
	or	al, al
	loopnz	p2_copy
	pop	si
pi_no_file:
	call	parse_cline
pi_exit_ok:
	clc
pi_exit:
	pop	si
pi_ret:
	mov	di, (bp+VAR_ENV_TAIL)
	mov	al, #0
	stosb
!
! ok, now that the environment is build, declare its presence to
! the kernel.
!
	mov	CL_OFFSET, #BUFF_BOOT_ENV
	mov	CL_MAGIC_ADDR, #CL_MAGIC
	ret

!
! Apply simple rule whether to put a leading space before copied
! word or not.
!
leading_space:
	mov	di, (bp+VAR_ENV_TAIL)
leading_space_2di:
	cmp	di, #BUFF_BOOT_ENV
	jz	no_leading_space
	mov	al, #0x20
	cmp	(di-1), byte #'=
	jz	no_leading_space
	cmp	al, (di-1)
	jz	no_leading_space
	stosb
no_leading_space:
	ret


is_white_space:
	or	al, al		! NULL character terminates string
	jz	it_is
	cmp	al, #0x20	! SPACE
	jz	it_is
	cmp	al, #0x0a	! NL
	jz	it_is
	cmp	al, #0x09	! TAB
it_is:	ret


!
! compute word length in CX
!
word_length:
	push	si
	xor	cx, cx
char_next:
	lodsb
	call	is_white_space
	jz	str_fin
	inc	cx
	jmp	char_next
str_fin:
	pop	si
	ret

!
! Locate '=' sign in the current word pointed to by SI. This job gets
! done by first getting the string length.
!
locate_eqs:
	call	word_length	! sets CX to word length
locate_cx_eqs:
	push	di
	push	cx
	mov	di, si
	mov	al, #'=
	repnz
	scasb			! locate the end of the string
	mov	bx, di
	pop	cx		! return with meaningfull CX
	pop	di
	ret

!
! lookup configuration filename in GETS_INPUT area.
!
lookup_config:
	lea	si, (bp+VAR_GETS_INPUT)
	cmp	(si), byte #0
	jnz	this_word
config_n_found:
		mov	si, #msg_config
		ret
this_word:
	call	locate_eqs		! lookup the variable by name.
	jz	eqs____found
		cmp	(bx), byte #0	! test whether the terminating
		jz	config_n_found	! character was NULL
lc_next_word:
		inc	cx		! skip the trailing SPACE
		add	si, cx		! CX == word length
		jmp	this_word
eqs____found:
	mov	di, #msg_var_config
	xchg	cx, bx			! after, BX == word length, CX == '=' position
	sub	cx, si			! CX == '=' offset
	mov	dx, cx			! save varname length
	push	si			! save word origin
	repz
	cmpsb
	jz	lupc_found
		pop	si		! word origin
		mov	cx, bx		! restore word length
		jmp	lc_next_word
lupc_found:
	pop	ax			! discard word origin value
	sub	bx, dx			! update word length (options only)
	mov	di, #BUFF_INODE1
	push	di			! save pathname destination
	mov	cx, bx
	rep
	movsb
	mov	al, #0
	stosb				! copy the terminating NULL, too.
	pop	si			! return with pointer to destination
	ret


/*
 * parse command line syntax, which is:
 *	[filename] [var=value] [-option]
 *	?
 *	? dirname
 * i.e. words not beginning with a dach '-' are treated as
 * kernel filenames. words containing equal sign are passed
 * as environement variables.
*/
parse_cline:
	call	leading_space
next_spc:
	lodsb
	or	al, al		! NULL ends parsing
	jnz	not_exit
		mov	(bp+VAR_ENV_TAIL), di
		ret
not_exit:
	cmp	al, #0x20	! SPACE
	jz	next_spc	! skip spaces

	cmp	al, #'-
	jnz	not_opt
	call	word_concat_2di
	jmp	next_spc
not_opt:
	push	ax
	call	locate_eqs	! test the presance of '=' sign
	pop	ax
	jnz	not_var
	dec	si		! (ungetchar)
	call	word_concat_2di
	jmp	next_spc
not_var:			! if nothing else, kernel filename
	push	si
	mov	si, #msg_var_image
	call	word_concat_2di
	pop	si
	dec	si		! (ungetchar)
	call	word_concat_2di
	jmp	next_spc

!
! parse single line of boot configuration file using command line
! syntax. This is different from the cline syntax in that only
! NL terminate words; SPACEs and TABs are removed.
!
!  config syntax is:
!	[var=value]
!	# is a comment
!
! YUP, I know; I should allow for line continuation somehow...
! ... not in this release, though.
!
parse_fline:
	call	leading_space
	xor	ax, ax
	cmp	(si), byte #'#
	jnz	next_b		! this is a comment line --> exit here
pf_exit:
		mov	(bp+VAR_ENV_TAIL), di
		ret
next_b:
	lodsb
	or	al, al		! NULL means finish
	jz	pf_exit
	cmp	al, #0x0a	! NL means finish as well
	jz	pf_exit
	cmp	al, #0x20	! SPACEs ...
	jz	next_b		! ... are silently skipped
	cmp	al, #0x09	! ... same goes for TABs
	jz	next_b
	stosb			! everything else gets copyed
	jmp	next_b

!
! the following two entry points allow for word concatenation
! at the end of BUFF_BOOT_ENV area. 
!
word_concat:
	mov	di, (bp+VAR_ENV_TAIL)
word_concat_2di:
	call	leading_space_2di
no_spc_here:
	lodsb
	cmp	al, #0x20		! strings cannot contain SPACEs
	jz	end_str
	or	al, al
	jz	end_str
	cmp	di, #BUFF_BOOT_ENV_END
	jnc	end_str
	stosb
	jmp	no_spc_here
end_str:
	dec	si			! unget  the terminating character...
	mov	(bp+VAR_ENV_TAIL),  di
	ret

!
! a procedure to lookup env variable in BOOT_ENV *backword*.
! variable name is pointed by SI, together with a default
! value which followes it (separated only by a single NULL
! byte).
!
hunt_backword:
	mov	(bp+VAR_SEARCH_TARGET), si
	mov	di, (bp+VAR_ENV_TAIL)
	mov	cx, di
	sub	cx, #BUFF_BOOT_ENV
search_next_eqs:
	jcxz	variable_not_found
	mov	si, (bp+VAR_SEARCH_TARGET)
	mov	al, #'=
	std
	repnz
	scasb
	jcxz	variable_not_found

	mov	al, #0x20
	mov	bx, di
	inc	bx
	repnz
	scasb
	push	di
	jcxz	only_one_char	! we ended up at the begin of the
	inc	di		! buffer; we *must* adjust pointer.
	inc	di
only_one_char:
	push	cx
	mov	cx, bx
	sub	cx, di
	inc	cx
	cld			! I should do case insensitive comparision
	repz			! here. But,... may be next release...
	cmpsb
	mov	si, di
	pop	cx
	pop	di
	jnz	search_next_eqs
	lea	di, (bp+VAR_GETS_INPUT)
	push	di			! save pointer to  the returned string
copy_value:
	lodsb
	stosb
	or	al, al
	jz	end_copy_value
	cmp	al, #0x20
	jnz	copy_value
	mov	(di-1), byte #0		! overwrite  the terminating SPACE
end_copy_value:
	pop	si			! SI points to the returned string
hb_exit:
	or	si, si
	ret
variable_not_found:
	xor	si, si
	jmp	hb_exit

!
! Locates BOOTIMAGE and extracts image name
!
! Returns with SI pointing to GET_INPUT where image
! filename ends up. The search must start at the tail
! of the string, so that the last definition takes
! precedence.
!
lookup_bootimage:
	mov	si, #msg_var_image
	call	hunt_backword		! lookup the variable by name.
	jz	image_not_found
		ret
image_not_found:
	cld
	mov	si, (bp+VAR_SEARCH_TARGET) ! #msg_var_image
	call	word_concat		! sets DI
	inc	si			! YES!! the variable name and its
					! default value *have* to be separated
					! by exactly one byte (i.e NULL)
	push	si
	call	word_concat_2di		! does not set DI
	mov	(di), byte #0
	mov	(bp+VAR_ENV_TAIL),  di
	pop	si
	ret

!
! locate 'ramdisk=' env variable and set RAMDISK accordingly.
!
lookup_ramdisk:
	mov	si, #msg_var_ramdisk
	call	hunt_backword		! lookup the variable by name.
	jnz	ramdisk_found
		ret
ramdisk_found:
	call	atoi
	mov	504, ax
	ret

!
! locate 'svga=' env variable and set VGAMODE accordingly.
!
lookup_svga:
	mov	506, word #-1		! SVGA mode

	mov	si, #msg_var_svga
	call	hunt_backword		! lookup the variable by name.
	jnz	svga_found
		ret
svga_found:
	call	atoi
	mov	506, ax
	ret

!
! convert ascii text pointed to by SI into a numerical value
!
atoi:
	xor	cx, cx
	xor	ax, ax
	mov	bx, #10
	push	si
	cmp	(si), #'-
	jnz	next_digit
	inc	si
next_digit:
	lodsb
	sub	al, #'0
	jc	atoi_end
	cmp	al, #9
	ja	atoi_end
	xchg	ax, cx
	mul	bx
	add	ax, cx
	xchg	ax, cx
	jmp	next_digit
atoi_end:
	xchg	ax, cx
	pop	si
	cmp	(si), #'-
	jnz	atoi_exit
	neg	ax
atoi_exit:
	ret


!
! scanns command line prepared for the booted kernel;
! locates BOOTFLAGS and splits  them into separate words.
!
lookup_bootflags:
	mov	si, #BUFF_BOOT_ENV
	mov	cx, (bp+VAR_ENV_TAIL)
	sub	cx, si
lookup_again:
	push	cx		! CX == length of the remaining ENV area
	call	locate_eqs
	jz	eq_found
goto_next_word:
	inc	cx		! skip the trailing SPACE
	add	si, cx		! CX == word length
	pop	ax
	xchg	ax, cx
	sub	cx, ax
	ja	lookup_again
		ret
eq_found:
	mov	di, #msg_var_bflags
	xchg	cx, bx		! after, BX == word length, CX == '=' position
	sub	cx, si		! CX == '=' offset
	mov	dx, cx		! save varname length
	push	si
	repz
	cmpsb
	jz	bf_found
		pop	si	! word origin
no_more_comas:
		mov	cx, bx
		jmp	goto_next_word
bf_found:
! 1) remove VARNAME
	pop	ax		! word origin
	xchg	ax, di
	pop	cx		! recall remaining ENV size
	sub	cx, dx		! reduce it by the length of varname
	push	cx		! put the updated value back
	sub	bx, dx		! update word length (options only)
	push	di		! save new options origin
	rep
	movsb
	mov	(bp+VAR_ENV_TAIL),  di
	movsb			! copy the terminating NULL, too.
	pop	si
! 2) overwrite comas by spaces (in current word being the value of
! BOOTFLAGS environment var)
	mov	di, si
	mov	al, #',
	mov	cx, bx		! get the options string length
punch_again:
	repnz
	scasb
	jnz	no_more_comas	! ---> goto  the next word
	mov	(di-1), byte #0x20
	jmp	punch_again


!
! gets single line from opened file.
!
fgets:
	lea	di, (bp+VAR_GETS_INPUT)
	mov	cx, (bp+VAR_BUFF_REMAIN)
	mov	si, (bp+VAR_FILE_PTR)
	or	cx, cx
	jnz	begin_scan
do_reload:
	push	di
	call	reload_iobuff
	pop	di
	mov	si, bx			! 'reload_iobuff' sets BX to BUFF_BLOCK
	stc
	jcxz	end_of_file
begin_scan:
	lodsb
	stosb
	cmp	al, #0x0a		! i.e. NewLine
	jz	end_of_line
	loop	begin_scan
	jmp	do_reload		! BUFF_BLOCK have ended --> reload
end_of_line:
	dec	cx
	clc
end_of_file:
	mov	(bp+VAR_FILE_PTR), si
	mov	(bp+VAR_BUFF_REMAIN), cx
	mov	al, #0
	stosb
	lea	si, (bp+VAR_GETS_INPUT)
	ret


reload_iobuff:
	mov	bx, #BUFF_BLOCK
	mov	eax, (bp+VAR_FILE_CURBLOCK)
	cmp	eax, (bp+VAR_CURRENT_INODE+EXT2IN_I_BLOCKS)
	jc	ok_load
	xor	cx, cx
	ret
ok_load:
	mov	cx, #SIZEOF_FS_BLOCK	! this much we'll read (at most)
	mov	di, bx
	xor	ax, ax
	rep
	stosb				! start with erasing the buffer
	call	file_block
	or	eax, eax		! if this block is not allocated
	jz	update_file_data	! don't call block loader
	call	load_block
update_file_data:
	mov	cx, #SIZEOF_FS_BLOCK
	inc	(bp+VAR_FILE_CURBLOCK)
	mov	eax, (bp+VAR_FILE_CURBLOCK)
	cmp	eax, (bp+VAR_CURRENT_INODE+EXT2IN_I_BLOCKS)
	jc	relo_exit
	mov	cx, (bp+VAR_CURRENT_INODE+EXT2IN_I_SIZE)
	and	cx, #SIZEOF_FS_BLOCK-1
relo_exit:
	ret


/**************** functions for string manipulation *****************/

#if 0
!
! locate string DI in string SI
!
hunt_forward:
	lodsb
	or	al, al
	jnz	not_ends
		ret
not_ends:
	call	is_white_space
	jz	hunt_forward
	dec	si
	call	locate_eqs
	jnz	not_present
! yes, there is an equal sign. compare prefixes..
		ret
not_present:
	add	si, cx
	jmp	hunt_forward
#endif
!
! locates string DI in string SI (size of SI is CX)
!
locate_string:
	call	strncmp
	ret

!
! compare semi caseinsensitive way. i.e. ES:DI is assumed to
! be lower case only.
!
strncmp:
	push	si
main_loop:
	lodsb
	or	al, al
	jz	snc_exit
	cmp	al, #0x20
	jc	sc_less
	cmp	al, #0x80
	jnc	sc_less
	or	al, #0x20
sc_less:
	scasb
	jz	main_loop
snc_exit:
	pop	si
	ret

!
! set DI and CX registers acording to string pointed to
! by SI. (this is not a true STRLEN(), CX includes the 
! terminating NULL)
! 
!
strlen:
	xor	ax,ax
	mov	cx, #-1
	mov	di, si
	cld
	repnz
	scasb
	not	cx
	sub	di, #2		! di pointed _after_ the terminating 0
	ret

!
! locate character AL in string SI. on return SI is
! unchanged, DI points to the end of the string.
!
strchr:
	push	ax
	call	strlen
	pop	ax
strnchr:
	mov	di, si
	cld
	repnz
	scasb
	ret


/**************** functions parsing filesystem ****************/

follow_link:
!
! test if this is a fast link; if so load pointer and fall
! through to 'build_name_stack'.
!
	lea	si, (bp+VAR_CURRENT_INODE+EXT2IN_BLOCK)
	cmp	(bp+VAR_CURRENT_INODE+EXT2IN_I_BLOCKS), word #0
	jz	fast_link
	!
	! Not a fast link! Load file block, it does contain the
	! link information. The assumption is that a path is at
	! most 1024 bytes long, so a symbolic link is contained
	! in a single fs block.
	!
		mov	eax, (si)
		mov	bx, #BUFF_BLOCK
		call	load_block
		mov	si, bx
fast_link:
	call	strlen
!
! Fall through to....
!
push_path_names_on_stack:
	push	si		! save begin of path
	xchg	di, si
	mov	di, (bp+VAR_NSTACK)
	std
nxt_name:
	xor	ax, ax
in_name:
	stosb
!	cmp	(di+1), word #0
!	jnz	ppn_skip
!	inc	di
ppn_skip:
	lodsb
	cmp	al, #'/
	loopnz	in_name
	jcxz	ppnos
	jmp	nxt_name
ppnos:	pop	si
	mov	(bp+VAR_NSTACK), di
	cld
	ret

!
! call get_inode for a file with path pointed to by SI
!
! This routine is called recursively, although it does not build
! up a return stack. The recursion here builds a name stack in 
! VAR_BUFF_NAME_STACK -- it finishes when this stack is 100%
! consumed.
!
! VAR_NSTACK point to *one* character *before* the first name
! begins.
!
fopen:
	xor	eax, eax
	mov	(bp+VAR_FILE_CURBLOCK), eax
	mov	(bp+VAR_FILE_PTR), ax
	mov	(bp+VAR_BUFF_REMAIN), ax
!
! fall through into the 'open_root_path' procedure....
!
open_root_path:
!
! at this point we start from ROOT_DIR inode
!
	db	0x66
	mov	(bp+VAR_CURDIR_INODE), #EXT2_ROOT_INO
	dw	0		! store 32-bit value...
!
! at this point we start at current work directory (whatever that
! is).
!
open_path:
	call	strlen		! set registers

	mov	(bp+VAR_NSTACK), word #PTR_NSTACK_INIT	! name stack pointer
	call	push_path_names_on_stack

into_dir:

	mov	eax, (bp+VAR_CURDIR_INODE)
	call	get_inode

	mov	(bp+VAR_DIR_ACTION), word #name_lookup
	call	process_directory
	jz	ld_failed	! -> name not found in this directory
	!
	! Found! open the file
	!
		mov	(bp+VAR_NSTACK), di
		push	di
		call	get_inode
		pop	di
	!
	! first test whether we found a symbolic link here
	!
		cmp	(bp+VAR_CURRENT_INODE), #0xa1ff
		jnz	not_symlink
			call	follow_link
			jmp	into_dir
	not_symlink:
	!
	! ...NO. It's neither a directory nor an ordinary file.
	! In either case we can finish right now.
	!
		cmp	di, #PTR_NSTACK_INIT
		jnc	ld_exit		! no more names on the
					! VAR_BUFF_NAME_STACK -> exit.
					! Carry is reset to signal
					! success.
	!
	! check whether the file is a directory and whether we
	! have more names to lookup. If it's not a directory we
	! failed, for there are more names in the path while
	! we found an ordinary file in the middle of it.
	!
		test	(bp+VAR_CURRENT_INODE), #0x04000
		jz	ld_failed	! ZERO == not a dir
			mov	eax, (bp+VAR_CUROPN_INODE)
			mov	(bp+VAR_CURDIR_INODE), eax
			jmp	into_dir
ld_failed:
	stc
ld_exit:
	ret

!
! print directory contents (first test whether the opened file
! is a directory).
!
print_dir:
	mov	(bp+VAR_DIR_ACTION), word #print_fname
	test	(bp+VAR_CURRENT_INODE), #0x04000
	jnz	process_directory
		ret			! nothing to do!!
!
! fall through into the 'process_directory' !!!
! the following procedure processes a file as if it was
! a directory; calling (VAR_DIR_ACTION) on every directory entry
!
process_directory:
	xor	eax, eax
next_block:
		push	eax
		call	file_block
		or	eax, eax
		jz	end_lookup
		mov	bx, #BUFF_BLOCK
		call	load_block
indir:
			mov	ax, (bp+VAR_DIR_ACTION)
			call	ax
			jnz	dir_nfound
				or	eax, eax ! NZ == something found
				jmp	end_lookup
dir_nfound:
			cmp	bx, #BUFF_BLOCK+SIZEOF_FS_BLOCK
			jc	indir
block_parsed:
		pop	eax
		inc	eax
	cmp	eax, (bp+VAR_CURRENT_INODE+EXT2IN_I_BLOCKS)
	jbe	next_block
	xor	ax, ax		! signal ZERO
end_lookup:
	pop	ebx		! discard stack item
	ret


!
! compare one name on stack with the current dir entry. return
! inode if they match. Otherwise return 0. BX gets adjusted
! to the next directory entry.
!
name_lookup:
	push	si
	mov	di, (bp+VAR_NSTACK)
skip_nuls:
	cmp	di, #PTR_NSTACK_INIT	! we are at the end --> this means
	jc	do_lookup		! that skipping initial nulls we
err_exit:				! have consumed pathname --> this
	or	di, di			! means that no path was requested
	jmp	nl_exit			! --> return with ZERO flag reset
do_lookup:
	inc	di
	cmp	(di), byte #0		! '//' path element ...
	jz	skip_nuls
	mov	eax, (bx+EXT2_DIR_INODE)
	mov	cx, (bx+EXT2_DIR_NAMLEN)
	lea	si, (bx+EXT2_DIR_NAME)
	add	bx, (bx+EXT2_DIR_RECLEN)
	repz
	cmpsb
	jnz	nl_exit
	cmp	(di), byte #0		! ... and the NULL at the end of string
nl_exit:
	pop	si
	clc			! CF == 0 --> directory (not a file)
	ret

!
! gets a string from the console...
!
gets:	push	di
	lea	di, (bp+VAR_GETS_INPUT)
gets_1: xor	ax, ax
	int	0x16
	or	al, al
	jz	gets_1
	cmp	al, #0x08
	jnz	gets_2
	dec	di
	call	putc
	push	ax
	mov	al, #0x20
	call	putc
	pop	ax
	jmp	gets_3
gets_2:
	stosb
gets_3:	call	putc
	cmp	al, #0x0d
	jnz	gets_1
	dec	di		! drop  the terminating '\n'
	mov	al, #0x0a
	call	putc		! echo NewLine
	xor	ax, ax
	stosb
	pop	di
	ret

!
! print filename from a ext2fs directory entry pointed to by
! DS:BX, followed by CR LF sequence. BX gets adjusted by the
!  size of direntry.
!
print_fname:
	push	si
	mov	cx, (bx+EXT2_DIR_NAMLEN)
	lea	si, (bx+EXT2_DIR_NAME)
	add	bx, (bx+EXT2_DIR_RECLEN)
print_sicx:
	lodsb
	mov	ah, #0x0e
	int	0x10
	loop	print_sicx
	pop	si
print_crlf:
	mov	al, #0x0d
	call	putc
	mov	al, #0x0a
putc:
	mov	ah, #0x0e
	int	0x10
	or	ax, ax		! clean zero bit
	clc			! CF == 0 --> directory (not a file)
	ret

load_indirect1:
	push	bx
	sub	ax, #LOC_IND_1
	push	eax		! save updated index
	mov	eax, (si+LOC_IND_01)
	call	load_ind1
	pop	eax
	mov	ecx, #4000
	xor	edx, edx
	div	ecx
	push	dx
	mov	si, ax
	mov	eax, (bx+si)
	call	load_indirect0
	pop	si
	mov	eax, (bx+si)
	pop	bx		! the load pointer (not control)
	ret

!
! support routine
!
load_ind1:
	mov	bx, #BUFF_INODE1
	cmp	eax, (bp+VAR_INODE1_CACHE)
	jz	li_ret
	mov	(bp+VAR_INODE1_CACHE), eax
	push	es
	push	ss
	pop	es
	call	load_block
	pop	es
li_ret:	ret


puts:
	lodsb
	or	al, al
	jnz	not_nul
	ret
not_nul:
	cmp	al, #0x0a
	jnz	not_nl
	mov	ax, #0x0e0d
	int	0x10
	mov	al, #0x0a
	jmp	not_tab
not_nl:
	cmp	al, #9
	jnz	not_tab
	mov	al, #0x20
not_tab:
	mov	ah, #0x0e
	int	0x10
	jmp	puts

/*
 * This procedure turns off the floppy drive motor, so
 * that we enter the kernel in a known state, and
 * don't have to worry about it later.
 */
kill_motor:
	push dx
	mov dx,#0x3f2
	xor al, al
	outb
	pop dx
	ret


msg_nofile:
	.byte	0x0d, 0x0a
	.ascii	"No such file."
	.byte	0
msg_boot:
	.byte	0x0d, 0x0a
	.ascii	"Linux boot loader. Type ? for help."
	.byte	0x0d, 0x0a
	.byte	0x0d, 0x0a
	.ascii	"boot: "
	.byte	0
msg_config_from:
	.byte	0x0d, 0x0a
	.ascii	"Configuration from: "
	.byte	0
msg_booting:
	.byte	0x0d, 0x0a
	.ascii	"Booting file: "
	.byte	0
msg_help_file:
	.ascii	"/unix/kernel/boot.help"
	.byte	0
msg_var_ramdisk:
	.ascii	"ramdisk="
!	.byte	0
msg_ramdisk:
	.ascii	"0"
	.byte	0
msg_var_svga:
	.ascii	"svga="
!	.byte	0
msg_svga:
	.ascii	"-1"
	.byte	0
msg_var_config:
	.ascii	"config="
!	.byte	0
msg_config:
	.ascii	"/etc/boot.conf"
	.byte	0
msg_var_bflags:
	.ascii	"BOOTFLAGS="
msg_var_image:
	.ascii	"BOOTIMAGE="
	.byte	0
msg_image:
	.ascii	"/linux_kernel"
	.byte	0
msg_auto:
	.ascii	"auto"
	.byte	0

strptr_image:
	.word	0

	end
